@@ -7,10 +7,10 @@ import random
 from compas.geometry import Frame
 from compas.geometry import Sphere
 from compas.geometry import Transformation
+from compas.robots import Configuration
 from compas.robots import Joint
 from compas.robots import RobotModel
 
-from compas_fab.robots.configuration import Configuration
 from compas_fab.robots.constraints import Constraint
 from compas_fab.robots.constraints import JointConstraint
 from compas_fab.robots.constraints import OrientationConstraint
@@ -55,6 +55,10 @@ class Robot(object):
         self.semantics = semantics
         self.client = client
         self.attributes = {}
+        self._current_ik = {
+            'request_id': None,
+            'solutions': None
+        }
 
     @property
     def artist(self):
@@ -211,8 +215,7 @@ class Robot(object):
         """
         if not full_configuration:
             full_configuration = self.zero_configuration()
-        full_joint_state = dict(zip(full_configuration.joint_names, full_configuration.values))
-        return self.model.forward_kinematics(full_joint_state, link_name=self.get_end_effector_link_name(group))
+        return self.model.forward_kinematics(full_configuration, link_name=self.get_end_effector_link_name(group))
 
     def get_base_link_name(self, group=None):
         """Get the name of the robot's base link.
@@ -274,8 +277,7 @@ class Robot(object):
         """
         if not full_configuration:
             full_configuration = self.zero_configuration()
-        full_joint_state = dict(zip(full_configuration.joint_names, full_configuration.values))
-        return self.model.forward_kinematics(full_joint_state, link_name=self.get_base_link_name(group))
+        return self.model.forward_kinematics(full_configuration, link_name=self.get_base_link_name(group))
 
     def get_link_names(self, group=None):
         """Get the names of the links in the kinematic chain.
@@ -359,7 +361,7 @@ class Robot(object):
         :obj:`list` of :attr:`compas.robots.Joint.SUPPORTED_TYPES`
             List of joint types.
         """
-        return [self.get_joint_by_name(n).type for n in names]
+        return self.model.get_joint_types_by_names(names)
 
     def get_joint_by_name(self, name):
         """RGet the joint in the robot model matching the given name.
@@ -464,7 +466,7 @@ class Robot(object):
 
         Returns
         -------
-        :obj:`list` of :obj:`float`
+        :class:`compas.robots.Configuration`
 
         Note
         ----
@@ -497,9 +499,8 @@ class Robot(object):
             The configuration of the group.
         """
         full_configuration = self._check_full_configuration_and_scale(full_configuration)[0]  # adds joint_names to full_configuration and makes copy
-        full_joint_state = dict(zip(full_configuration.joint_names, full_configuration.values))
         group_joint_names = self.get_configurable_joint_names(group)
-        values = [full_joint_state[name] for name in group_joint_names]
+        values = [full_configuration[name] for name in group_joint_names]
         return Configuration(values, self.get_configurable_joint_types(group), group_joint_names)
 
     def merge_group_with_full_configuration(self, group_configuration, full_configuration, group):
@@ -533,7 +534,7 @@ class Robot(object):
 
         full_configuration = self._check_full_configuration_and_scale(full_configuration)[0]  # adds joint_names to full_configuration and makes copy
 
-        full_configuration.merge(group_configuration)
+        full_configuration = full_configuration.merged(group_configuration)
         return full_configuration
 
     def get_group_names_from_link_name(self, link_name):
@@ -579,10 +580,10 @@ class Robot(object):
             match the configurable joints of the given `group`.
         """
         names = self.get_configurable_joint_names(group)
-        if len(names) != len(configuration.values):
+        if len(names) != len(configuration.joint_values):
             raise ValueError(
-                "Please pass a configuration with %d values or specify group" % len(names))
-        return configuration.values[names.index(joint_name)]
+                "Please pass a configuration with %d joint_values or specify group" % len(names))
+        return configuration.joint_values[names.index(joint_name)]
 
     def _check_full_configuration_and_scale(self, full_configuration=None):
         """Either create a full configuration or check if the passed full configuration is valid.
@@ -602,7 +603,7 @@ class Robot(object):
         else:
             joint_names = self.get_configurable_joint_names()  # full configuration
             # full_configuration might have passive joints specified as well, we allow this.
-            if len(joint_names) > len(full_configuration.values):
+            if len(joint_names) > len(full_configuration.joint_values):
                 raise ValueError("Please pass a configuration with {} values, for all configurable joints of the robot.".format(len(joint_names)))
             configuration = full_configuration.copy()
             if not configuration.joint_names:
@@ -866,7 +867,7 @@ class Robot(object):
        """
         if not self.client:
             raise Exception(
-                'This method is only callable once a client is assigned')
+                'This method is only callable once a client is assigned.')
 
     def ensure_semantics(self):
         """Check if semantics is set.
@@ -877,8 +878,16 @@ class Robot(object):
         """
         if not self.semantics:
             raise Exception(
-                'This method is only callable once a semantic model is assigned')
+                'This method is only callable once a semantic model is assigned.')
 
+    def ensure_geometry(self):
+        """Check if the model's geometry has been loaded.
+        Raises
+        ------
+        :exc:`Exception`
+            If geometry has not been loaded.
+        """
+        self.model.ensure_geometry()
     # ==========================================================================
     # constraints
     # ==========================================================================
@@ -1088,22 +1097,22 @@ class Robot(object):
             group = self.main_group_name
 
         joint_names = self.get_configurable_joint_names(group)
-        if len(joint_names) != len(configuration.values):
-            raise ValueError("The passed configuration has %d values, the group %s needs however: %d" % (
-                len(configuration.values), group, len(joint_names)))
+        if len(joint_names) != len(configuration.joint_values):
+            raise ValueError("The passed configuration has %d joint_values, the group %s needs however: %d" % (
+                len(configuration.joint_values), group, len(joint_names)))
         if len(tolerances_above) == 1:
             tolerances_above = tolerances_above * len(joint_names)
-        elif len(tolerances_above) != len(configuration.values):
-            raise ValueError("The passed configuration has %d values, the tolerances_above however: %d" % (
-                len(configuration.values), len(tolerances_above)))
+        elif len(tolerances_above) != len(configuration.joint_values):
+            raise ValueError("The passed configuration has %d joint_values, the tolerances_above however: %d" % (
+                len(configuration.joint_values), len(tolerances_above)))
         if len(tolerances_below) == 1:
             tolerances_below = tolerances_below * len(joint_names)
-        elif len(tolerances_below) != len(configuration.values):
-            raise ValueError("The passed configuration has %d values, the tolerances_below however: %d" % (
-                len(configuration.values), len(tolerances_below)))
+        elif len(tolerances_below) != len(configuration.joint_values):
+            raise ValueError("The passed configuration has %d joint_values, the tolerances_below however: %d" % (
+                len(configuration.joint_values), len(tolerances_below)))
 
         constraints = []
-        for name, value, tolerance_above, tolerance_below in zip(joint_names, configuration.values, tolerances_above, tolerances_below):
+        for name, value, tolerance_above, tolerance_below in zip(joint_names, configuration.joint_values, tolerances_above, tolerances_below):
             constraints.append(JointConstraint(name, value, tolerance_above, tolerance_below))
         return constraints
 
@@ -1114,6 +1123,15 @@ class Robot(object):
     def inverse_kinematics(self, frame_WCF, start_configuration=None, group=None, return_full_configuration=False, options=None):
         """Calculate the robot's inverse kinematic for a given frame.
 
+        The inverse kinematic solvers are implemented as generators in order to fit both analytic
+        and numerical solver approaches. However, this method abstracts that away and returns one
+        configuration at a time to simplify its usage.
+
+        To keep the usefulness of the generator, calls to this method will recall the last retrieved
+        iterator and keep returning results yielded by it, one at a time. This is true only as long as
+        the request is identical to the last one. If the arguments change, the last generator
+        is discarded and the client IK implementation is invoked again.
+
         Parameters
         ----------
         frame_WCF : :class:`compas.geometry.Frame`
@@ -1141,7 +1159,7 @@ class Robot(object):
         Returns
         -------
         :class:`Configuration`
-            The planning group's configuration.
+            An inverse kinematic solution represented as a configuration.
 
         Examples
         --------
@@ -1151,8 +1169,68 @@ class Robot(object):
         >>> robot.inverse_kinematics(frame_WCF, start_configuration, group)                 # doctest: +SKIP
         Configuration((4.045, 5.130, -2.174, -6.098, -5.616, 6.283), (0, 0, 0, 0, 0, 0))    # doctest: +SKIP
         """
+        # Pseudo-memoized sequential calls will re-use iterator if not exhaused
+        request_id = '{}-{}-{}-{}-{}'.format(str(frame_WCF), str(start_configuration), str(group), str(return_full_configuration), str(options))
+
+        if self._current_ik['request_id'] == request_id and self._current_ik['solutions'] is not None:
+            solution = next(self._current_ik['solutions'], None)
+            if solution is not None:
+                return solution
+
+        solutions = self.iter_inverse_kinematics(frame_WCF, start_configuration, group, return_full_configuration, options)
+        self._current_ik['request_id'] = request_id
+        self._current_ik['solutions'] = solutions
+
+        return next(solutions)
+
+    def iter_inverse_kinematics(self, frame_WCF, start_configuration=None, group=None, return_full_configuration=False, options=None):
+        """Iterate over the inverse kinematic solutions of a robot.
+
+        This method exposes the generator-based inverse kinematic solvers. Analytics solvers will return
+        generators that include all possible solutions, hence exhausting the iterator indicates there are
+        no more solutions to be found. Numerical solvers, on the other hand, will keep returning solutions
+        until one fails to be found (usually caused by a timeout rather than infeasibility) or until the user
+        code stops the iteration.
+
+        Parameters
+        ----------
+        frame_WCF : :class:`compas.geometry.Frame`
+            The frame to calculate the inverse kinematic for.
+        start_configuration : :class:`Configuration`, optional
+            If passed, the inverse will be calculated such that the calculated
+            joint positions differ the least from the start_configuration.
+            Defaults to the zero configuration.
+        group: :obj:`str`, optional
+            The planning group used for calculation. Defaults to the robot's
+            main planning group.
+        return_full_configuration : :obj:`bool`, optional
+            If ``True``, returns a full configuration with all joint values
+            specified, including passive ones if available. Defaults to ``False``.
+        options: :obj:`dict`, optional
+            Dictionary containing the key-value pairs of additional options.
+            The valid options are specific to the backend in use.
+            Check the API reference of the IK backend implementation for more details.
+
+        Raises
+        ------
+        :exc:`compas_fab.backends.BackendError`
+            If no configuration can be found.
+
+        Yields
+        -------
+        :class:`Configuration`
+            An inverse kinematic solution represented as a configuration.
+
+        Examples
+        --------
+        >>> frame_WCF = Frame([0.3, 0.1, 0.5], [1, 0, 0], [0, 1, 0])
+        >>> start_configuration = robot.zero_configuration()
+        >>> group = robot.main_group_name
+        >>> next(robot.iter_inverse_kinematics(frame_WCF, start_configuration, group))      # doctest: +SKIP
+        Configuration((4.045, 5.130, -2.174, -6.098, -5.616, 6.283), (0, 0, 0, 0, 0, 0))    # doctest: +SKIP
+        """
         options = options or {}
-        attached_collision_meshes = options.get('attached_collision_meshes', [])
+        attached_collision_meshes = options.get('attached_collision_meshes') or []
 
         self.ensure_client()
         group = group or self.main_group_name if self.semantics else None
@@ -1167,12 +1245,20 @@ class Robot(object):
 
         options['attached_collision_meshes'] = attached_collision_meshes
 
+        solutions = self.client.inverse_kinematics(self,
+                                                   frame_WCF_scaled,
+                                                   start_configuration_scaled,
+                                                   group,
+                                                   options)
+
         # The returned joint names might be more than the requested ones if there are passive joints present
-        joint_positions, joint_names = self.client.inverse_kinematics(self,
-                                                                      frame_WCF_scaled,
-                                                                      start_configuration_scaled,
-                                                                      group,
-                                                                      options)
+        for joint_positions, joint_names in solutions:
+            if joint_positions:
+                yield self._build_configuration(joint_positions, joint_names, group, return_full_configuration)
+            else:
+                yield None  # to accomodate analytic ik with keeping the order of solutions
+
+    def _build_configuration(self, joint_positions, joint_names, group, return_full_configuration):
         if return_full_configuration:
             # build configuration including passive joints, but no sorting
             joint_types = self.get_joint_types_by_names(joint_names)
@@ -1262,8 +1348,6 @@ class Robot(object):
         full_configuration = self.merge_group_with_full_configuration(configuration, self.zero_configuration(), group)
         full_configuration, full_configuration_scaled = self._check_full_configuration_and_scale(full_configuration)
 
-        full_joint_state = dict(zip(full_configuration.joint_names, full_configuration.values))
-
         # If there's no client, we default to `model` solver if there is no other assigned.
         if not self.client:
             solver = solver or 'model'
@@ -1275,7 +1359,7 @@ class Robot(object):
             if link not in self.get_link_names(group):
                 raise ValueError('Link name {} does not exist in planning group'.format(link))
 
-            frame_WCF = self.model.forward_kinematics(full_joint_state, link)
+            frame_WCF = self.model.forward_kinematics(full_configuration, link)
 
         # Otherwise, pass everything down to the client
         else:
@@ -1351,7 +1435,7 @@ class Robot(object):
         options = options or {}
         max_step = options.get('max_step')
         path_constraints = options.get('path_constraints')
-        attached_collision_meshes = options.get('attached_collision_meshes', [])
+        attached_collision_meshes = options.get('attached_collision_meshes') or []
 
         self.ensure_client()
         if not group:
@@ -1481,8 +1565,8 @@ class Robot(object):
         >>> robot.client = RosClient()
         >>> robot.client.run()
         >>> configuration = Configuration.from_revolute_values([0.0, -1.5707, 0.0, -1.5707, 0.0, 0.0])
-        >>> tolerances_above = [math.radians(5)] * len(configuration.values)
-        >>> tolerances_below = [math.radians(5)] * len(configuration.values)
+        >>> tolerances_above = [math.radians(5)] * len(configuration.joint_values)
+        >>> tolerances_below = [math.radians(5)] * len(configuration.joint_values)
         >>> group = robot.main_group_name
         >>> goal_constraints = robot.constraints_from_configuration(configuration, tolerances_above, tolerances_below, group)
         >>> trajectory = robot.plan_motion(goal_constraints, start_configuration, group, {'planner_id': 'RRTConnectkConfigDefault'})
@@ -1494,7 +1578,7 @@ class Robot(object):
         """
         options = options or {}
         path_constraints = options.get('path_constraints')
-        attached_collision_meshes = options.get('attached_collision_meshes', [])
+        attached_collision_meshes = options.get('attached_collision_meshes') or []
 
         # TODO: for the motion plan request a list of possible goal constraints
         # can be passed, from which the planner will try to find a path that
@@ -1596,8 +1680,7 @@ class Robot(object):
         """
         if not len(configuration.joint_names):
             configuration.joint_names = self.get_configurable_joint_names(group)
-        joint_state = dict(zip(configuration.joint_names, configuration.values))
-        return self.model.transformed_frames(joint_state)
+        return self.model.transformed_frames(configuration)
 
     def transformed_axes(self, configuration, group=None):
         """Get the robot's transformed axes.
@@ -1617,8 +1700,7 @@ class Robot(object):
         """
         if not len(configuration.joint_names):
             configuration.joint_names = self.get_configurable_joint_names(group)
-        joint_state = dict(zip(configuration.joint_names, configuration.values))
-        return self.model.transformed_axes(joint_state)
+        return self.model.transformed_axes(configuration)
 
     # ==========================================================================
     # drawing
@@ -1645,9 +1727,8 @@ class Robot(object):
 
         if not len(configuration.joint_names):
             configuration.joint_names = self.get_configurable_joint_names(group)
-        joint_state = dict(zip(configuration.joint_names, configuration.values))
 
-        self.artist.update(joint_state, visual, collision)
+        self.artist.update(configuration, visual, collision)
 
     def draw_visual(self):
         """Draw the robot's visual geometry using the defined :attr:`Robot.artist`."""
